import { checkStrictlyIncreasing, trimPoly, evaluatePolySegment } from "./Utils";
var EPSILON = Number.EPSILON || 2.2204460492503130808472633361816E-16;
export function createAkimaSplineInterpolator(xVals, yVals) {
    var segmentCoeffs = computeAkimaPolyCoefficients(xVals, yVals);
    var xValsCopy = Float64Array.from(xVals);
    return function (x) { return evaluatePolySegment(xValsCopy, segmentCoeffs, x); };
}
export function computeAkimaPolyCoefficients(xVals, yVals) {
    if (xVals.length != yVals.length) {
        throw new Error("Dimension mismatch for xVals and yVals.");
    }
    if (xVals.length < 5) {
        throw new Error("Number of points is too small.");
    }
    checkStrictlyIncreasing(xVals);
    var n = xVals.length - 1;
    var differences = new Float64Array(n);
    var weights = new Float64Array(n);
    for (var i = 0; i < n; i++) {
        differences[i] = (yVals[i + 1] - yVals[i]) / (xVals[i + 1] - xVals[i]);
    }
    for (var i = 1; i < n; i++) {
        weights[i] = Math.abs(differences[i] - differences[i - 1]);
    }
    var firstDerivatives = new Float64Array(n + 1);
    for (var i = 2; i < n - 1; i++) {
        var wP = weights[i + 1];
        var wM = weights[i - 1];
        if (Math.abs(wP) < EPSILON && Math.abs(wM) < EPSILON) {
            var xv = xVals[i];
            var xvP = xVals[i + 1];
            var xvM = xVals[i - 1];
            firstDerivatives[i] = (((xvP - xv) * differences[i - 1]) + ((xv - xvM) * differences[i])) / (xvP - xvM);
        }
        else {
            firstDerivatives[i] = ((wP * differences[i - 1]) + (wM * differences[i])) / (wP + wM);
        }
    }
    firstDerivatives[0] = differentiateThreePoint(xVals, yVals, 0, 0, 1, 2);
    firstDerivatives[1] = differentiateThreePoint(xVals, yVals, 1, 0, 1, 2);
    firstDerivatives[n - 1] = differentiateThreePoint(xVals, yVals, n - 1, n - 2, n - 1, n);
    firstDerivatives[n] = differentiateThreePoint(xVals, yVals, n, n - 2, n - 1, n);
    return computeHermitePolyCoefficients(xVals, yVals, firstDerivatives);
}
function differentiateThreePoint(xVals, yVals, indexOfDifferentiation, indexOfFirstSample, indexOfSecondsample, indexOfThirdSample) {
    var x0 = yVals[indexOfFirstSample];
    var x1 = yVals[indexOfSecondsample];
    var x2 = yVals[indexOfThirdSample];
    var t = xVals[indexOfDifferentiation] - xVals[indexOfFirstSample];
    var t1 = xVals[indexOfSecondsample] - xVals[indexOfFirstSample];
    var t2 = xVals[indexOfThirdSample] - xVals[indexOfFirstSample];
    var a = (x2 - x0 - (t2 / t1 * (x1 - x0))) / (t2 * t2 - t1 * t2);
    var b = (x1 - x0 - a * t1 * t1) / t1;
    return (2 * a * t) + b;
}
function computeHermitePolyCoefficients(xVals, yVals, firstDerivatives) {
    if (xVals.length != yVals.length || xVals.length != firstDerivatives.length) {
        throw new Error("Dimension mismatch");
    }
    if (xVals.length < 2) {
        throw new Error("Not enough points.");
    }
    var n = xVals.length - 1;
    var segmentCoeffs = new Array(n);
    for (var i = 0; i < n; i++) {
        var w = xVals[i + 1] - xVals[i];
        var w2 = w * w;
        var yv = yVals[i];
        var yvP = yVals[i + 1];
        var fd = firstDerivatives[i];
        var fdP = firstDerivatives[i + 1];
        var coeffs = new Float64Array(4);
        coeffs[0] = yv;
        coeffs[1] = firstDerivatives[i];
        coeffs[2] = (3 * (yvP - yv) / w - 2 * fd - fdP) / w;
        coeffs[3] = (2 * (yv - yvP) / w + fd + fdP) / w2;
        segmentCoeffs[i] = trimPoly(coeffs);
    }
    return segmentCoeffs;
}
//# sourceMappingURL=Akima.js.map