import { checkMonotonicallyIncreasing, checkFinite, getMedian } from "./Utils";
import { createInterpolatorWithFallback } from "./Index";
export function createLoessInterpolator(parms) {
    var _a = parms.interpolationMethod, interpolationMethod = _a === void 0 ? "akima" : _a, _b = parms.minXDistance, minXDistance = _b === void 0 ? getDefaultMinXDistance(parms.xVals) : _b, diagInfo = parms.diagInfo;
    var fitYVals = smooth(parms);
    var knotFilter = createKnotFilter(parms.xVals, fitYVals, minXDistance);
    var knotXVals = filterNumberArray(parms.xVals, knotFilter);
    var knotYVals = filterNumberArray(fitYVals, knotFilter);
    if (diagInfo) {
        diagInfo.fitYVals = fitYVals;
        diagInfo.knotFilter = knotFilter;
        diagInfo.knotXVals = knotXVals;
        diagInfo.knotYVals = knotYVals;
    }
    return createInterpolatorWithFallback(interpolationMethod, knotXVals, knotYVals);
}
function createKnotFilter(xVals, fitYVals, minXDistance) {
    var n = xVals.length;
    var filter = Array(n);
    var prevX = -Infinity;
    for (var i = 0; i < n; i++) {
        var x = xVals[i];
        var y = fitYVals[i];
        if (x - prevX >= minXDistance && !isNaN(y)) {
            filter[i] = true;
            prevX = x;
        }
        else {
            filter[i] = false;
        }
    }
    return filter;
}
function filterNumberArray(a, filter) {
    var n = a.length;
    var a2 = new Float64Array(n);
    var n2 = 0;
    for (var i = 0; i < n; i++) {
        if (filter[i]) {
            a2[n2++] = a[i];
        }
    }
    return a2.subarray(0, n2);
}
function getDefaultMinXDistance(xVals) {
    var n = xVals.length;
    if (n == 0) {
        return NaN;
    }
    var xRange = xVals[n - 1] - xVals[0];
    if (xRange == 0) {
        return 1;
    }
    return xRange / 100;
}
export function smooth(parms) {
    var xVals = parms.xVals, yVals = parms.yVals, weights = parms.weights, _a = parms.bandwidthFraction, bandwidthFraction = _a === void 0 ? 0.3 : _a, _b = parms.robustnessIters, robustnessIters = _b === void 0 ? 2 : _b, _c = parms.accuracy, accuracy = _c === void 0 ? 1E-12 : _c, _d = parms.outlierDistanceFactor, outlierDistanceFactor = _d === void 0 ? 6 : _d, diagInfo = parms.diagInfo;
    checkMonotonicallyIncreasing(xVals);
    checkFinite(yVals);
    if (weights) {
        checkFinite(weights);
    }
    var n = xVals.length;
    if (yVals.length != n || weights && weights.length != n) {
        throw new Error("Dimension mismatch.");
    }
    if (diagInfo) {
        diagInfo.robustnessIters = 0;
        diagInfo.secondLastMedianResidual = undefined;
        diagInfo.lastMedianResidual = undefined;
        diagInfo.robustnessWeights = undefined;
    }
    if (n <= 2) {
        return Float64Array.from(yVals);
    }
    var fitYVals = undefined;
    for (var iter = 0; iter <= robustnessIters; iter++) {
        var robustnessWeights = undefined;
        if (iter > 0) {
            var residuals = absDiff(fitYVals, yVals);
            var medianResidual = getMedian(residuals);
            if (medianResidual < accuracy) {
                if (diagInfo) {
                    diagInfo.lastMedianResidual = medianResidual;
                }
                break;
            }
            var outlierDistance = medianResidual * outlierDistanceFactor;
            robustnessWeights = calculateRobustnessWeights(residuals, outlierDistance);
            if (diagInfo) {
                diagInfo.robustnessIters = iter;
                diagInfo.secondLastMedianResidual = medianResidual;
                diagInfo.robustnessWeights = robustnessWeights;
            }
        }
        var combinedWeights = combineWeights(weights, robustnessWeights);
        fitYVals = calculateSequenceRegression(xVals, yVals, combinedWeights, bandwidthFraction, accuracy, iter);
    }
    return fitYVals;
}
function calculateSequenceRegression(xVals, yVals, weights, bandwidthFraction, accuracy, iter) {
    var n = xVals.length;
    var n2 = weights ? countNonZeros(weights) : n;
    if (n2 < 2) {
        throw new Error("Not enough relevant points in iteration " + iter + ".");
    }
    var bandwidthInPoints = Math.max(2, Math.min(n2, Math.round(n2 * bandwidthFraction)));
    var bw = findInitialBandwidthInterval(weights, bandwidthInPoints, n);
    var fitYVals = new Float64Array(n);
    for (var i = 0; i < n; i++) {
        var x = xVals[i];
        moveBandwidthInterval(bw, x, xVals, weights);
        fitYVals[i] = calculateLocalLinearRegression(xVals, yVals, weights, x, bw.iLeft, bw.iRight, accuracy);
    }
    return fitYVals;
}
export function calculateLocalLinearRegression(xVals, yVals, weights, x, iLeft, iRight, accuracy) {
    var maxDist = Math.max(x - xVals[iLeft], xVals[iRight] - x) * 1.001;
    if (maxDist < 0) {
        throw new Error("Inconsistent bandwidth parameters.");
    }
    if (maxDist == 0) {
        maxDist = 1;
    }
    var sumWeights = 0;
    var sumX = 0;
    var sumXSquared = 0;
    var sumY = 0;
    var sumXY = 0;
    for (var k = iLeft; k <= iRight; ++k) {
        var xk = xVals[k];
        var yk = yVals[k];
        var dist = Math.abs(xk - x);
        var w1 = weights ? weights[k] : 1;
        var w2 = triCube(dist / maxDist);
        var w = w1 * w2;
        var xkw = xk * w;
        sumWeights += w;
        sumX += xkw;
        sumXSquared += xk * xkw;
        sumY += yk * w;
        sumXY += yk * xkw;
    }
    if (sumWeights < 1E-12) {
        return NaN;
    }
    var meanX = sumX / sumWeights;
    var meanY = sumY / sumWeights;
    var meanXY = sumXY / sumWeights;
    var meanXSquared = sumXSquared / sumWeights;
    var meanXSqrDiff = meanXSquared - meanX * meanX;
    var beta;
    if (Math.abs(meanXSqrDiff) < Math.pow(accuracy, 2)) {
        beta = 0;
    }
    else {
        beta = (meanXY - meanX * meanY) / meanXSqrDiff;
    }
    return meanY + beta * x - beta * meanX;
}
function findInitialBandwidthInterval(weights, bandwidthInPoints, n) {
    var iLeft = findNonZero(weights, 0);
    if (iLeft >= n) {
        throw new Error("Initial bandwidth start point not found.");
    }
    var iRight = iLeft;
    for (var i = 0; i < bandwidthInPoints - 1; i++) {
        iRight = findNonZero(weights, iRight + 1);
        if (iRight >= n) {
            throw new Error("Initial bandwidth end point not found.");
        }
    }
    return { iLeft: iLeft, iRight: iRight };
}
function moveBandwidthInterval(bw, x, xVals, weights) {
    var n = xVals.length;
    while (true) {
        var nextRight = findNonZero(weights, bw.iRight + 1);
        if (nextRight >= n || xVals[nextRight] - x >= x - xVals[bw.iLeft]) {
            return;
        }
        bw.iLeft = findNonZero(weights, bw.iLeft + 1);
        bw.iRight = nextRight;
    }
}
function calculateRobustnessWeights(residuals, outlierDistance) {
    var n = residuals.length;
    var robustnessWeights = new Float64Array(n);
    for (var i = 0; i < n; i++) {
        robustnessWeights[i] = biWeight(residuals[i] / outlierDistance);
    }
    return robustnessWeights;
}
function combineWeights(w1, w2) {
    if (!w1 || !w2) {
        return w1 || w2;
    }
    var n = w1.length;
    var a = new Float64Array(n);
    for (var i = 0; i < n; i++) {
        a[i] = w1[i] * w2[i];
    }
    return a;
}
function findNonZero(a, startPos) {
    if (!a) {
        return startPos;
    }
    var n = a.length;
    var i = startPos;
    while (i < n && a[i] == 0) {
        i++;
    }
    return i;
}
function countNonZeros(a) {
    var cnt = 0;
    for (var i = 0; i < a.length; i++) {
        if (a[i] != 0) {
            cnt++;
        }
    }
    return cnt;
}
function absDiff(a1, a2) {
    var n = a1.length;
    var a3 = new Float64Array(n);
    for (var i = 0; i < n; i++) {
        a3[i] = Math.abs(a1[i] - a2[i]);
    }
    return a3;
}
function triCube(x) {
    var absX = Math.abs(x);
    if (absX >= 1) {
        return 0;
    }
    var tmp = 1 - absX * absX * absX;
    return tmp * tmp * tmp;
}
function biWeight(x) {
    var absX = Math.abs(x);
    if (absX >= 1) {
        return 0;
    }
    var tmp = 1 - absX * absX;
    return tmp * tmp;
}
//# sourceMappingURL=Loess.js.map